/*-----------------------------------------------------------------*/
/*! 
  \file scscpfileclient.cxx
  \brief SCSCP client over file or socket
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008, 2009, M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr

   \bug M. GASTINEAU 20/04/09 : update for SCSCP 1.3
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This library is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#if HAVE_WINDOWS_H
#include <windows.h>
#endif /*HAVE_CONFIG_H*/
#include <sys/types.h>
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETDB_H
#include <netdb.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdarg.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "scscptypes.h"
#include "scscpinternal.h"
#include "scscpfileclient.h"
#include "scscpfileserver.h"
#include "scscpprocedurecall.h"
#include "scscpdebug.h"
#define __SCSCP_WITHIN_SCSCP
#include "scscp.h"

/*-----------------------------------------------------------------*/
/* private defines */
/*-----------------------------------------------------------------*/
/*! convert xmlchar to char */
#define xmlChar2char(x) ((char*)(x))

/*-----------------------------------------------------------------*/
/* private types */
/*-----------------------------------------------------------------*/
/*! paramter for hook excute functions */
typedef struct {
 const char *cdname;
 const char *symbolname;
 int (*callbackwriteargs)(SCSCP_io* stream, void *param, SCSCP_status* status);
 void *param;
} t_paramexecutehook;

/*-----------------------------------------------------------------*/
/* private functions */
/*-----------------------------------------------------------------*/
static int SCSCP_sc_callsendstr_callback(SCSCP_io* client, void *param, SCSCP_status* status);
static int SCSCP_sc_clearallowedversions(SCSCP_Fileclient* client, SCSCP_status* localstatus);

/*-----------------------------------------------------------------*/
/*! connect to a server at the specified port
    return 1 on success
    on error, geterror() will give the error message
    @param servername (in) name of the server
    @param port (in) port on the server of the SCSCP server
    @param status (out) status
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_connectsocket(SCSCP_Fileclient* client, const char *servername, int port, SCSCP_status* status)
{
  int bres  = 1;
  static int binitsocket=1;
  const char *supportedversion;
  struct hostent * hp;
  SCSCP_debugprint("SCSCP_Fileclient::connect( %s, %d) - entree\n",servername,port);
  
  if (binitsocket && !SCSCP_initsocket())
  {
   SCSCP_debugprint("SCSCP_Fileclient::connect() - SCSCP_initsocket failed\n");
   SCSCP_status_seterror(status, SCSCP_STATUS_ERRNO);
   bres = 0;
  }
  
  hp = gethostbyname(servername);
  if ( hp == NULL ) 
  {
   SCSCP_debugprint("SCSCP_Fileclient::connect() - gethostbyname failed\n");
   SCSCP_status_seterror(status, SCSCP_STATUS_ERRNO);
   bres = 0;
  }
  /* creer une socket pour se connecter au serveur */
  else if ((client->m_fd = socket(AF_INET, SOCK_STREAM, 0))==INVALID_SOCKET)
  {
   SCSCP_debugprint("SCSCP_Fileclient::connect() - socket failed\n");
   SCSCP_status_seterror(status, SCSCP_STATUS_ERRNO);
   bres = 0;
  }
  else
  {
   struct sockaddr_in adresseServeur;
   adresseServeur.sin_family = AF_INET;
   adresseServeur.sin_port = htons(port);/* host to network short */
   memcpy(&adresseServeur.sin_addr.s_addr, hp->h_addr, hp->h_length);

   SCSCP_debugprint("SCSCP_Fileclient::connect() - socket sucess m_fd=%d\n", client->m_fd);
   SCSCP_debugprint("SCSCP_Fileclient::connect() - server address : %s\n", inet_ntoa(adresseServeur.sin_addr));
   /* se connecter au serveur */
   if (connect(client->m_fd, (struct sockaddr *)&adresseServeur, sizeof(adresseServeur))== SOCKET_ERROR ) 
   {
    SCSCP_debugprint("SCSCP_Fileclient::connect() - connect failed\n");
    SCSCP_status_seterror(status, SCSCP_STATUS_ERRNO);
    bres = 0;
    closesocket(client->m_fd);  
    client->m_fd = -1; 
   }
   else
   {
    SCSCP_fileio_setfile(&client->m_OutStream,client->m_fd, 1);
    SCSCP_fileio_setfile(&client->m_InStream,client->m_fd, 1);
    bres = SCSCP_sc_readCIM(client, &supportedversion, status);
    if (bres) bres = SCSCP_sc_writeCIMVersion(client, supportedversion);
    if (bres) bres = SCSCP_sc_readCIMVersion(client, status);
    if (!bres) SCSCP_sc_closesocket(client);
   }
  }
  binitsocket = 0;
  SCSCP_debugprint("SCSCP_Fileclient::connect(this=%p, m_fd = %d) - returns %d\n", client, client->m_fd, bres);
  return bres;
}


/*-----------------------------------------------------------------*/
//! close the connection to the server
/*-----------------------------------------------------------------*/
void SCSCP_sc_closesocket(SCSCP_Fileclient* client)
{
 SCSCP_debugprint("SCSCP_Fileclient::Close() - entree\n");
 if (client->m_fd!=-1) 
 { 
  const char *msg="<?scscp quit ?>\n";
  SCSCP_fileio_write(&client->m_OutStream, msg, strlen(msg)*sizeof(char));
  closesocket(client->m_fd);  
  client->m_fd = -1; 
 }
 SCSCP_debugprint("SCSCP_Fileclient::Close() - sortie\n");
}

/*-----------------------------------------------------------------*/
/*! read a line from the input stream
  and return 1 on success
 @param buffer (out) buffer to fill
 @param lenbuffer (in) size of buffer
 @param status (inout) error code
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_readline(SCSCP_Fileclient* client, char *buffer, int lenbuffer, SCSCP_status *status)
{
    int bres = 1;
    int j=-1;
    
    SCSCP_debugprint("SCSCP_Fileclient::ReadLine() - entree\n");
    
    do
    {
     j++;
     bres=(SCSCP_fileio_read(&client->m_InStream, buffer+j, sizeof(char))!=SOCKET_ERROR);
    } while (bres && buffer[j]!='\n' && j<lenbuffer);
    if (j==lenbuffer) bres = 0;
    if (!bres)
    {
     SCSCP_status_seterror(status, SCSCP_STATUS_RECVQUIT);
    }
    else
    {
     buffer[j] = '\0';
     SCSCP_debugprint("SCSCP_Fileclient::ReadLine - read '%s' j=%d\n", buffer, j);
    }
    
    SCSCP_debugprint("SCSCP_Fileclient::ReadLine - sortie et retourne %d\n", bres);
    return bres;
}
/*-----------------------------------------------------------------*/
/*! read the Connection information Message from the server
  and return 1 on success
 @param supportedversion (out) supported version
  @param status (inout) status 
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_readCIM(SCSCP_Fileclient* client, const char **supportedversion,SCSCP_status* status)
{
    int bres;
    char c[SCSCP_PI_MAXLENBUFFER];
    char *scscp_versions;
    char *lasts;
    
    SCSCP_debugprint("SCSCP_Fileclient::ReadCIM() - entree\n");
 
     bres = SCSCP_sc_readline(client,c, SCSCP_PI_MAXLENBUFFER, status);
     *supportedversion = NULL;
     if (bres)
     {
      scscp_versions = SCSCP_findPIstring(c,"scscp_versions");
      if (scscp_versions)
      {
         char *vers;
         for (vers=strtok_r(scscp_versions, " ", &lasts); vers && *supportedversion==NULL; vers= strtok_r(NULL, " ", &lasts))
         {
          size_t j;
          for ( j=0; j<client->m_allowedversionscount && *supportedversion==NULL; j++)
          {
            if (strcmp(vers,client->m_allowedversions[j])==0) *supportedversion = client->m_allowedversions[j];
          }
         }
         /* aucune version supporter => on donne la + grande */
         if (*supportedversion==NULL) *supportedversion=client->m_allowedversions[0];
      }
      else
      {
       bres = 0;
       SCSCP_status_seterror(status, SCSCP_STATUS_VERSIONNEGOTIATIONFAILED);
      }
    }
    
    SCSCP_debugprint("SCSCP_Fileclient::ReadCIM - sortie et retourne %d\n", bres);
    return bres;
}


/*-----------------------------------------------------------------*/
/*! read the Connection information Message Version from the server
  and return 1 on success
  @param status (inout) status 
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_readCIMVersion(SCSCP_Fileclient* client, SCSCP_status* status)
{
    int bres;
    char c[SCSCP_PI_MAXLENBUFFER];
    char *scscp_versions;
    
    SCSCP_debugprint("SCSCP_Fileclient::ReadCIMVersion() - entree\n");
 
     bres = SCSCP_sc_readline(client,c,  SCSCP_PI_MAXLENBUFFER, status);
     if (bres)
     {
      scscp_versions = SCSCP_findPIstring(c,"version");
      if (scscp_versions==NULL) 
      {
       bres = 0;
       SCSCP_status_seterror(status, SCSCP_STATUS_VERSIONNEGOTIATIONFAILED);
      }
    }
    
    SCSCP_debugprint("SCSCP_Fileclient::ReadCIMVersion - sortie et retourne %d\n", bres);
    return bres;
}

/*-----------------------------------------------------------------*/
/*! write the SCSCP version of the client to the server
  and return 1 on success
 @param supportedversion (in) supported version
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_writeCIMVersion(SCSCP_Fileclient* client, const char *supportedversion)
{
    int bres;
     const char *msg1="<?scscp version=\"";
     const char *msg2="\" ?>\n";
     
    SCSCP_debugprint("SCSCP_Fileclient::WriteCIMVersion() - entree\n");
 
    bres =  SCSCP_fileio_write(&client->m_OutStream,msg1, strlen(msg1)*sizeof(char));
    if (bres) bres = SCSCP_fileio_write(&client->m_OutStream,supportedversion, strlen(supportedversion)*sizeof(char));
    if (bres) bres = SCSCP_fileio_write(&client->m_OutStream,msg2, strlen(msg2)*sizeof(char));
     
    SCSCP_debugprint("SCSCP_Fileclient::WriteCIMVersion - sortie et retourne %d\n", bres);
    return bres;
}

/*-----------------------------------------------------------------*/
/*! attach the specified to the client
    @param socket (in) socket to be attached
*/
/*-----------------------------------------------------------------*/
void SCSCP_sc_setsocket(SCSCP_Fileclient* client,int socket)
{
 client->m_fd = socket;
 SCSCP_fileio_setfile(&client->m_OutStream,client->m_fd, 1);
 SCSCP_fileio_setfile(&client->m_InStream, client->m_fd, 1);
}

/*-----------------------------------------------------------------*/
/*! checks if SCSCP client is null and returns 0 on error
  @param client (in) object to be checked
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_checkNULL(SCSCP_socketclient* client, SCSCP_status* status)
{
   int res = 1;
   
   if (client==NULL || *client==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_CLIENTOBJECTNULL);
     res = 0;
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! initialize the object client and set status if an error occurs 
 @param client (inout) object to be initialized
 @param status (inout) status error
 xparam ...    (in) allowed version to be negotiated
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_init(SCSCP_socketclient* client, SCSCP_status* status, ...)
{
   va_list versions;
   int res = 1;

   SCSCP_debugprint("SCSCP_sc_init(%p, %s) - enter\n", client,  status);
   
   if (client==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_CLIENTOBJECTNULL);
     res = 0;
   }
   else
   {
    SCSCP_Fileclient* pclient;
    *client = pclient = ( SCSCP_Fileclient*)SCSCP_malloc(sizeof(SCSCP_Fileclient), status);
    if (pclient==NULL) 
    {
     res = 0;
    }
    else
    {
     /* count elements */
     size_t  counter = 0;
     va_list listcount;
     
     
     va_start(listcount, status);
     while (va_arg(listcount, const char*)!=NULL) counter++;
     va_end(listcount);

     pclient->m_doc =NULL;
     pclient->m_fd = -1;
     
     /* duplicate the elements of versions */
     pclient->m_allowedversions = (SCSCP_string*)SCSCP_malloc(sizeof(SCSCP_string)*counter, status);
     res = (pclient->m_allowedversions!=NULL);
     if (res)
     {
      size_t cpt;
      va_start(versions, status);
      for( cpt=0; cpt<counter; cpt++)
      {
       pclient->m_allowedversions[cpt] = SCSCP_strdup(va_arg(versions, const char*), status);
       res = (pclient->m_allowedversions[cpt]!=NULL);
      }
      pclient->m_allowedversionscount = counter;
      va_end(versions);
     }
     
    }
   }
   SCSCP_debugprint("SCSCP_sc_init(%p, %d) - return %d\n", *client,  SCSCP_status_is(status), res);
   return res;
}



/*-----------------------------------------------------------------*/
/*! clear the allowed version to be negiotiated to the client 
 @param client (inout) object to be cleared
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
static int SCSCP_sc_clearallowedversions(SCSCP_Fileclient* client, SCSCP_status* status)
{
   int res = 1;
   
   SCSCP_debugprint("SCSCP_sc_clearallowedversions() - enter\n");
   if (client==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_CLIENTOBJECTNULL);
     res = 0;
   }
   else
   {
    /* clear the elements of versions */
    if (client->m_allowedversions!=NULL)
    {
     size_t cpt;
     for(cpt=0; cpt<client->m_allowedversionscount; cpt++)
     {
      free(client->m_allowedversions[cpt]);
     }
     free(client->m_allowedversions);
     client->m_allowedversions = NULL;
     client->m_allowedversionscount = 0;
    }
   }
   SCSCP_debugprint("SCSCP_sc_clearallowedversions() - returns %d\n", res);
   return res;
}

/*-----------------------------------------------------------------*/
/*! clear the object client previously initialized by SCSCP_sc_init  
 @param client (inout) object to be deleted
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_clear(SCSCP_socketclient* client, SCSCP_status* status)
{
   int res;
   SCSCP_debugprint("SCSCP_sc_clear(%p,%p) - enter\n", client, status);
   
   res = SCSCP_sc_checkNULL(client, status);
   if (!res) return res;
   res = SCSCP_sc_clearallowedversions(*client, status);
   SCSCP_sc_closesocket(*client);
   SCSCP_sc_freexmldoc(*client);
   free (*client);
   
   SCSCP_debugprint("SCSCP_sc_clear(%d) - return %d\n", SCSCP_status_is(status), res);
   return res;
}

/*-----------------------------------------------------------------*/
/*! connect the client to a server on the specified host and port  
 @param client (inout) scscp client
 @param machine (in) computer name or IP address
 @param port (in) port on the remote computer
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_connect(SCSCP_socketclient* client, const char *machine, int port, SCSCP_status* status)
{
   int res;
   
   SCSCP_debugprint("SCSCP_sc_connect(%p, %s, %d, %p) - enter\n", client, machine, port, status);
   
   res = SCSCP_sc_checkNULL(client, status);
   if (!res) return res;
   
   res = SCSCP_sc_connectsocket(*client, machine, port,status);
   if (!res && SCSCP_status_is(status)==SCSCP_STATUS_OK) SCSCP_status_seterror(status,SCSCP_STATUS_ERRNO);

   SCSCP_debugprint("SCSCP_sc_connect(%d) - return %d\n", SCSCP_status_is(status), res);
   return res;
}
 
/*-----------------------------------------------------------------*/
/*! close the connection of the client to a server 
   previously opened by SCSCP_sc_connect  
 @param client (inout) scscp client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_close(SCSCP_socketclient* client, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (res)  SCSCP_sc_closesocket(*client);
   return res;
}

/*-----------------------------------------------------------------*/
/*! write the openmath buffer to the scscp server
 @param stream (inout) scscp stream
 @param param (in) openmath expression (must be a const char *)
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
static int SCSCP_sc_callsendstr_callback(SCSCP_io* stream, void *param, SCSCP_status* status)
{
  const char *openmathbuffer = (const char*)param;
  return SCSCP_io_write(stream, openmathbuffer, status);
}




/*-----------------------------------------------------------------*/
/*! send an openmath string to the SCSCP server with some options 
    for a procedure call
 @param client (inout) scscp client
 @param options (in) options of the call
 @param openmathbuffer (in) openmath expression
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_callsendstr(SCSCP_socketclient* client, SCSCP_calloptions* options, 
                          const char *openmathbuffer, SCSCP_status* status)
{
   return SCSCP_sc_callsendhook(client, options, SCSCP_sc_callsendstr_callback, (void*)openmathbuffer, status);
}

/*-----------------------------------------------------------------*/
/*! send a procedure call to the SCSCP server with some options 
    The parameter of the procedure call are written 
    by the callback function "callbackwriteargs"
 @param client (inout) scscp client
 @param options (in) options of the call
 @param callbackwriteargs (in) callback function 
             to write the parameters of the procedure call
 @param param (inout) opaque data given to callbackwriteargs
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_callsendhook(SCSCP_socketclient* client, SCSCP_calloptions* options, 
                          int (*callbackwriteargs)(SCSCP_io* stream, void *param, SCSCP_status* status),
                          void *param, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (res) res = SCSCP_procedurecall_write(client, options==NULL?NULL:*options, callbackwriteargs, param, status);
   return res;
}

/*-----------------------------------------------------------------*/
/*! the client sends the "information" message to the server. 
 @param client (inout) scscp client
 @param messagebuffer (in) information 
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_infomessagesend (SCSCP_socketclient* client, const char * messagebuffer, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (res) 
   {
     res = SCSCP_io_writef(client, status, "<?scscp info=\"", messagebuffer,"\" ?>\n", NULL);
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! the client sends the "interrupt" message to the server with the call id. 
 @param client (inout) scscp client
 @param call_id (in) call ID of the canceled procedure  
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_callsendinterrupt(SCSCP_socketclient* client, const char* call_id, SCSCP_status* status)
{
   int res = SCSCP_sc_checkNULL(client, status);
   if (res) 
   {
     res = SCSCP_io_writef(client, status, "<?scscp terminate call_id=\"", call_id,"\" ?>\n", NULL);
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! write the expression : 
   @param stream (inout) stream session
   @param status (inout) error type
   @param param (in) name of the remote object
*/
/*-----------------------------------------------------------------*/
static int SCSCP_sc_executehookwrite(SCSCP_io* stream, void *param, SCSCP_status *status)
{
 int res;
 t_paramexecutehook *hookparam = (t_paramexecutehook*)param;
 res = SCSCP_io_writeOMS(stream, hookparam->cdname, hookparam->symbolname, status);
 if (res)  res = hookparam->callbackwriteargs(stream, hookparam->param, status);
 return res;
}


/*-----------------------------------------------------------------*/
/*! send a procedure call to the server with symbolname and cdname and wait for the answer. 
 @param client (inout) scscp client
 @param returntype (in) return type
 @param cdname (inout) name of the CD
 @param symbolname (inout) name of the symbol
 @param callbackwriteargs (in) function to write arguments
 @param paramwrite (inout) opaque object given to callbackwriteargs
 @param node (out) xml node of the result
 @param status (inout) error code
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_executehookxmlnode(SCSCP_socketclient* client, SCSCP_option_return returntype, const char *cdname, const char* symbolname,
                              int (*callbackwriteargs)(SCSCP_io* stream, void *param, SCSCP_status* status), void *param, 
                              SCSCP_xmlnodeptr* node,
                              SCSCP_status* status)
{
 int res, res1;
 SCSCP_calloptions options;
 SCSCP_msgtype msgtype;
 char *cdnameerr;
 char *symbolnameerr;
 char *msg;
 
 t_paramexecutehook hookparam={cdname, symbolname, callbackwriteargs, param};
 
 SCSCP_debugprint("SCSCP_sc_executehookxmlnode(%p, %d, %s, %s, %p, %p, %p, %p)  - enter\n", client, (int)returntype, cdname, symbolname, callbackwriteargs , param, node, status);
 
 res1 = res = SCSCP_co_init(&options, status);
 if (res) res = SCSCP_co_set_returntype(&options, returntype,status);
 
 if (res) res = SCSCP_sc_callsendhook(client,&options, SCSCP_sc_executehookwrite, &hookparam, status);
 if (res) res = SCSCP_sc_callrecvheader(client, SCSCP_RETURNOPTIONS_IGNORE, &msgtype, status);
 if (res)
 {
  switch(msgtype)
  {
   case SCSCP_msgtype_ProcedureCompleted : 
           *node = SCSCP_sc_getxmlnode(client, status); 
           break;
           
   case SCSCP_msgtype_ProcedureTerminated : 
           *node = NULL;
           if (res) res = SCSCP_sc_callrecvterminated(client, &cdnameerr, &symbolnameerr, &msg, status);
           SCSCP_status_setexecfailed(status, cdnameerr, symbolnameerr, msg);
           res = 0;
           break;
           
   default : *node = NULL; break;
  }
 }

 if (res1) SCSCP_co_clear(&options, status);
 
 SCSCP_debugprint("SCSCP_sc_executehookxmlnode( %p, %d)  - returns %d\n", *node, SCSCP_status_is(status), res);
 return res;
}

/*-----------------------------------------------------------------*/
/*! send a procedure call to the server with symbolname and cdname and wait for the answer. 
 @param client (inout) scscp client
 @param returntype (in) return type
 @param cdname (inout) name of the CD
 @param symbolname (inout) name of the symbol
 @param callbackwriteargs (in) function to write arguments
 @param paramwrite (inout) opaque object given to callbackwriteargs
 @param openmathbuffer (out) openmath result
 @param status (inout) error code
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_executehookstr(SCSCP_socketclient* client, SCSCP_option_return returntype, const char *cdname, const char* symbolname,
                              int (*callbackwriteargs)(SCSCP_io* stream, void *param, SCSCP_status* status), void *param, 
                              char ** openmathbuffer,
                              SCSCP_status* status)
{
 int res, res1;
 SCSCP_calloptions options;
 SCSCP_msgtype msgtype;
 char *cdnameerr;
 char *symbolnameerr;
 char *msg;
 
 t_paramexecutehook hookparam={cdname, symbolname, callbackwriteargs, param};
 
 SCSCP_debugprint("SCSCP_sc_executehookstr(%p, %d, %s, %s, %p, %p, %p, %p)  - enter\n", client, (int)returntype, cdname, symbolname, callbackwriteargs , param, openmathbuffer, status);
 
 *openmathbuffer = NULL;
 res1 = res = SCSCP_co_init(&options, status);
 if (res) res = SCSCP_co_set_returntype(&options, returntype,status);
 
 if (res) res = SCSCP_sc_callsendhook(client,&options, SCSCP_sc_executehookwrite, &hookparam, status);
 if (res) res = SCSCP_sc_callrecvheader(client, SCSCP_RETURNOPTIONS_IGNORE, &msgtype, status);
 if (res)
 {
  switch(msgtype)
  {
   case SCSCP_msgtype_ProcedureCompleted : 
           res = SCSCP_sc_callrecvcompleted(client,openmathbuffer, status); 
           break;
           
   case SCSCP_msgtype_ProcedureTerminated : 
           if (res) res = SCSCP_sc_callrecvterminated(client, &cdnameerr, &symbolnameerr, &msg, status);
           SCSCP_status_setexecfailed(status, cdnameerr, symbolnameerr, msg);
           res = 0;
           break;
           
   default : *openmathbuffer = NULL; break;
  }
 }

 if (res1) SCSCP_co_clear(&options, status);
 
 SCSCP_debugprint("SCSCP_sc_executehook( %p, %d)  - returns %d\n", *openmathbuffer, SCSCP_status_is(status), res);
 return res;
}

/*-----------------------------------------------------------------*/
/*! return the first node in the current location of the openmath message
 @param client (inout) scscp client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
SCSCP_xmlnodeptr SCSCP_sc_getxmlnode(SCSCP_socketclient *client, SCSCP_status* status)
{
 int res = SCSCP_sc_checkNULL(client, status);
 return res?(*client)->m_iter:NULL;
}

/*-----------------------------------------------------------------*/
/*! \internal return the address of the first node in the current location of the openmath message
 @param client (inout) scscp client
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
SCSCP_xmlnodeptr* SCSCP_sc_getxmlnodeptr(SCSCP_socketclient *client, SCSCP_status* status)
{
 int res = SCSCP_sc_checkNULL(client, status);
 return res?&((*client)->m_iter):NULL;
}


/*----------------------------------------------------------------------------------------------*/
/*!  return node and its children as a raw string 
  return NULL  on error
  @param client (in) client. mustn't be NULL 
  @param node (in) current valid node. mustn't be NULL 
  @param status (in) error code
*/
/*----------------------------------------------------------------------------------------------*/
char* SCSCP_sc_getxmlnoderawstring(SCSCP_socketclient* client, SCSCP_xmlnodeptr node, SCSCP_status *status)
{
 char* str = NULL;
 int res = SCSCP_sc_checkNULL(client, status);
 xmlBufferPtr buffer;
 
 SCSCP_debugprint("SCSCP_sc_getxmlnoderawstring(%p,%p,%p) - enter\n",client, node, status);
 if (res==0) return str;
 
 buffer = xmlBufferCreate();
 if (xmlNodeDump(buffer, (*client)->m_doc, node, 1,1)!=-1)
 {
    const xmlChar *strxml = xmlBufferContent(buffer);
    if (strxml!=NULL)
    {
     int len = xmlStrlen(strxml)+1;
     str = (char*)SCSCP_malloc(sizeof(char)*len, status);
     if (str!=NULL)
     {
      strcpy(str, xmlChar2char(strxml));
     }
    }
    else
    {
     SCSCP_status_seterror(status, SCSCP_STATUS_NOMEM);
    }
 }
 else
 {
  SCSCP_status_seterror(status, SCSCP_STATUS_NOMEM);
 }
 
 SCSCP_debugprint(" SCSCP_sc_getxmlnoderawstring() - return '%s'\n", str);
 return str;
}

/*-----------------------------------------------------------------*/
/*! free the xml document
*/
/*-----------------------------------------------------------------*/
void SCSCP_sc_freexmldoc(SCSCP_Fileclient* client)
{
 if (client->m_doc) xmlFreeDoc(client->m_doc);
 client->m_doc = NULL;
}

/*-----------------------------------------------------------------*/
/* structure locale */
/*-----------------------------------------------------------------*/
typedef struct  {
/* t_controlpipe(SCSCP_fileio*  pfile, int pbfind, char *plastbuf, int plastbufvalid)
    : file(pfile), mathfind(pbfind), lastbufvalid(plastbufvalid)
 {
  status = SCSCP_STATUS_INITIALIZER;
  strcpy(lastbuf, plastbuf);
 };*/
 
 SCSCP_fileio*  file;      /*!< flux de type fichier */
 int mathfind ;  /*!< =1 => </math> trouve */
 char lastbuf[100]; /*!< dernier buffer lu */
 int lastbufvalid; /*!< lastbuf est valide */
 SCSCP_status status; /*!< status of the connection */
} t_controlpipe;


/*-----------------------------------------------------------------*/
/*! read the end of line
  @param fd (inout) file descriptor
  @param c (out) read buffer (at least 4094 bytes)
  @param status (inout) status
*/
/*-----------------------------------------------------------------*/
static int ReadLineBuffer(SCSCP_fileio* fd, char c[SCSCP_PI_MAXLENBUFFER], SCSCP_status* status)
{
    const char *symbolquit="<?scscp quit";
    int bres = 1;
    int j=-1;
    SCSCP_debugprint("ReadLineBuffer(%p,) - enter\n", fd);
    
    do
    {
     int ival;
     j++;
     ival = SCSCP_fileio_read(fd, c+j, sizeof(char));
     bres=(ival>0);
    } while (bres && c[j]!='\n' && j<=SCSCP_PI_MAXLENBUFFER-1);
    c[j]='\0';
    SCSCP_debugprint("ReadLineBuffer- read '%s'\n", c);
    
    if (strncmp(symbolquit, c,strlen(symbolquit))==0) 
    {
    SCSCP_debugprint("'<?scscp quit' found\n");
     SCSCP_status_seterror(status, SCSCP_STATUS_RECVQUIT);
    }
    SCSCP_debugprint("ReadLineBuffer(,%d)  -returns %d\n",SCSCP_status_is(status), bres);
    return bres;
}

/*-----------------------------------------------------------------------*/
/*! routine de lecture appelle par xmlReadIO.
 Doit etre compatible avec xmlInputReadCallback
 retourne la taille du buffer valide
 returne -1 lorsque </math> a deja ete trouve au precedent appel
 @param ctx (inout) contexte du flux
 @param buff (inout) buffer
 @param len (in) taille du buffer
*/
/*-----------------------------------------------------------------------*/
static int lecture(t_controlpipe* ctx, char* buff, int len)
{
    SCSCP_fileio*  file=ctx->file;
	int i=0;
	char *bufcmp;
	const char *symbol = "<?scscp";
	int lensymbol = strlen(symbol);
	int eof;
	
    if (!ctx->mathfind) 
    {
    do
    {
     errno=0;
     eof=(SCSCP_fileio_read(file, buff+i, sizeof(char))<=-1);
     /*printf("c=%c '%.*s'\n", buff[i], 7,(i>=7)?buff+i-7:"vide");*/
     /*SCSCP_debugprint(buff[i];*/
     i++;
     if (i<lensymbol && ctx->lastbufvalid)
     {
      memmove(ctx->lastbuf, ctx->lastbuf+1, (lensymbol-1)*sizeof(char));
      ctx->lastbuf[lensymbol-1] = buff[i-1];
     }
     else bufcmp = buff+i-lensymbol;
     
     if (i>=lensymbol) 
     {
        if (strncmp(bufcmp, symbol, lensymbol)==0)
        {
           char c[SCSCP_PI_MAXLENBUFFER];
           int bres;
           ctx->mathfind=1;
           i-=lensymbol;
           /*read the end of the line to cancel or not */
           bres = ReadLineBuffer(file, c, &ctx->status);
           if (!bres || strcmp(c," cancel ?>")==0) 
           {
              SCSCP_debugprint("cancel received\n");
              SCSCP_status_seterror(&ctx->status, SCSCP_STATUS_RECVCANCEL);
              i = -1;
           }
           if (strcmp(c," end ?>")==0) 
           {
              SCSCP_debugprint("end received\n");
           }
           break;
        }
     }
     
    } while (!eof && i<len );
 
     /* balise non trouvee : buffer plein */
     if (!ctx->mathfind)
     {
      if (i>=lensymbol) 
      {
       memcpy(ctx->lastbuf, buff+i-lensymbol, lensymbol*sizeof(char));
       ctx->lastbufvalid = 1;
      }
     }
    }
    else i=-1;
   
   if (i>0) SCSCP_debugprint("read from socket : '%.*s", i, buff);
   return i;
}



/*-----------------------------------------------------------------*/
/*! read the xml document. return 1 on success
  @param status (inout) error code
*/
/*-----------------------------------------------------------------*/
int SCSCP_sc_readxmldoc(SCSCP_Fileclient *client, SCSCP_status* status)
{
  int res;
  t_controlpipe ctx;
  xmlDocPtr doc;
  
  SCSCP_debugprint("\nSCSCP_sc_readxmldoc() - enter\n");
  
  LIBXML_TEST_VERSION
  
  ctx.file = &client->m_InStream;
  ctx.mathfind = 0;
  ctx.status = SCSCP_STATUS_INITIALIZER;
  strcpy(ctx.lastbuf, "                     ");
  ctx.lastbufvalid = 0;

  doc = xmlReadIO((xmlInputReadCallback)lecture,NULL, &ctx, NULL, NULL, XML_PARSE_NOERROR|XML_PARSE_NOWARNING| XML_PARSE_NOBLANKS);
  
  
  SCSCP_status_copy(status, &ctx.status);
  res = (doc!=NULL);
  if (res)
  {
   SCSCP_sc_freexmldoc(client);
   client->m_doc = doc;
  }
  else if (SCSCP_status_is(status)==SCSCP_STATUS_OK)
  {
   SCSCP_status_seterror(status, SCSCP_STATUS_OPENMATHNOTVALID);
  }
  
  SCSCP_debugprint("\nSCSCP_sc_readxmldoc(,%d) - returns %d\n", SCSCP_status_is(status), res);
  return res;
}

