/*-----------------------------------------------------------------*/
/*! 
  \file scscpio.cxx
  \brief SCSCP I/O function
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008, 2009, M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr

*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
 This library is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
*/
/*-----------------------------------------------------------------*/

#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#if HAVE_WINDOWS_H
#include <windows.h>
#endif /*HAVE_CONFIG_H*/
#include <sys/types.h>
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdarg.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "scscptypes.h"
#include "scscpinternal.h"
#include "scscpdebug.h"
#include "scscpfileclient.h"
#include "scscpfileserver.h"
#define __SCSCP_WITHIN_SCSCP
#include "scscp.h"
#include "float.h"

/*-----------------------------------------------------------------*/
/* private functions */
/*-----------------------------------------------------------------*/
static int SCSCP_io_checkNULL(SCSCP_io* stream, SCSCP_status* status);
static int SCSCP_io_flush(SCSCP_io* stream, SCSCP_status* status);

/*-----------------------------------------------------------------*/
/*! write some data to the SCSCP server or client 
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_write(SCSCP_io* stream, const char *buffer, SCSCP_status* status)
{
 int res = SCSCP_io_checkNULL(stream, status);
 if (res)
 {
   res = SCSCP_fileio_write(&(*stream)->m_OutStream, buffer, strlen(buffer));
  }
  return res;
}


/*-----------------------------------------------------------------*/
/*! flush the connection
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
static int SCSCP_io_flush(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream,"\n", status);
}

/*-----------------------------------------------------------------*/
/*! write some data to the SCSCP server or client 
 The list must finish a NULL pointer
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writef(SCSCP_io* stream,  SCSCP_status* status, const char *buffer, ...)
{
 va_list buflist;
 const char *data;
 int res = SCSCP_io_checkNULL(stream, status);
 if (res)
 {
   res = SCSCP_io_write(stream, buffer, status);
   
   va_start(buflist, buffer);
   while (res && ((data=va_arg(buflist, const char*))!=NULL)) 
   {
    res = SCSCP_io_write(stream, data, status);
   }
   va_end(buflist);
  }
  return res;
}

/*-----------------------------------------------------------------*/
/*! write buffer as an OpenMath string <OMSTR></OMSTR> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMSTR(SCSCP_io* stream, const char *buffer, SCSCP_status* status)
{
 int res;
 if (buffer!=NULL)
 {
  res =  SCSCP_io_writef(stream, status, "<OMSTR>", buffer, "</OMSTR>", NULL);
 }
 else
 {
  res =  SCSCP_io_write(stream, "<OMSTR/>", status);
 }
 return res;
}

/*-----------------------------------------------------------------*/
/*!  write x as an OpenMath float-point <OMF></OMF> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param x (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMFdouble(SCSCP_io* stream, double x, SCSCP_status* status)
{
 char valeur[400];
 sprintf(valeur,"<OMF dec=\"%+-.*E\" />",(int)DBL_DIG+1,x);
 return SCSCP_io_write(stream, valeur, status);
}

/*-----------------------------------------------------------------*/
/*! write buffer as an OpenMath floating-point <OMF></OMF> directly to the SCSCP client or server
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMFstr(SCSCP_io* stream, const char *buffer, SCSCP_status* status)
{
 return SCSCP_io_writef(stream, status, "<OMF dec=\"", buffer, "\" />", NULL);
}

/*-----------------------------------------------------------------*/
/*!   write x as an OpenMath integer <OMI></OMI> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param x (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMIint(SCSCP_io* stream, int x, SCSCP_status* status)
{
 char valeur[400];
 sprintf(valeur,"<OMI>%d</OMI>",x);
 return SCSCP_io_write(stream, valeur, status);
}

/*-----------------------------------------------------------------*/
/*!   write x as an OpenMath integer <OMI></OMI> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param x (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMIsizet(SCSCP_io* stream, size_t x, SCSCP_status* status)
{
 char valeur[400];
 sprintf(valeur,"<OMI>%llu</OMI>",(unsigned long long)x);
 return SCSCP_io_write(stream, valeur, status);
}

/*-----------------------------------------------------------------*/
/*! write x as an OpenMath integer <OMI></OMI> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param x (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMIlonglong(SCSCP_io* stream, long long  x, SCSCP_status* status)
{
 char valeur[400];
 sprintf(valeur,"<OMI>%lld</OMI>",x);
 return SCSCP_io_write(stream, valeur, status);
}

/*-----------------------------------------------------------------*/
/*! write buffer as an OpenMath integer <OMI></OMI> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMIstr(SCSCP_io* stream, const char *buffer, SCSCP_status* status)
{
 return SCSCP_io_writef(stream, status, "<OMI>", buffer, "</OMI>", NULL);
}

/*-----------------------------------------------------------------*/
/*! write symbolname/cdname  as an OpenMath symbol <OMS cd="..." name="..."> 
   directly to the SCSCP client or server
 return 0 on error
 @param client (inout) scscp stream
 @param cdname (in) cd name
 @param symbolname (in) symbol name
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMS(SCSCP_io* stream, const char* cdname, const char* symbolname, SCSCP_status* status)
{
 return SCSCP_io_writef(stream, status, "<OMS cd=\"", cdname, "\" name=\"", symbolname, "\" />", NULL);
}

/*-----------------------------------------------------------------*/
/*! write buffer as an OpenMath variable <OMV></OMV> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMV(SCSCP_io* stream, const char *buffer, SCSCP_status* status)
{
 return SCSCP_io_writef(stream, status, "<OMV>", buffer, "</OMV>", NULL);
}

/*-----------------------------------------------------------------*/
/*! write buffer as an OpenMath  reference <OMR xref="..."/> directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param buffer (in) some data
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMR(SCSCP_io* stream, const char *buffer, SCSCP_status* status)
{
 return SCSCP_io_writef(stream, status, "<OMR xref=\"", buffer, "\" />", NULL);
}

/*-----------------------------------------------------------------*/
/*! write the beginning OpenMath tag <OMA>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writebeginOMA(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "<OMA>", status);
}

/*-----------------------------------------------------------------*/
/*! write the end OpenMath tag </OMA>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeendOMA(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "</OMA>", status);
}

/*-----------------------------------------------------------------*/
/*! write the beginning OpenMath tag <OME>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writebeginOME(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "<OME>", status);
}

/*-----------------------------------------------------------------*/
/*! write the end OpenMath tag </OME>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeendOME(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "</OME>", status);
}


/*-----------------------------------------------------------------*/
/*! write the beginning OpenMath tag <OMOBJ>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writebeginOMOBJ(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "<OMOBJ>", status);
}

/*-----------------------------------------------------------------*/
/*! write the end OpenMath tag </OMOBJ>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeendOMOBJ(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "</OMOBJ>", status);
}

/*-----------------------------------------------------------------*/
/*! write the beginning OpenMath tag <OMATTR>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writebeginOMATTR(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "<OMATTR>", status);
}

/*-----------------------------------------------------------------*/
/*! write the end OpenMath tag </OMATTR>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeendOMATTR(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "</OMATTR>", status);
}

/*-----------------------------------------------------------------*/
/*! write the beginning OpenMath tag <OMATP>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writebeginOMATP(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "<OMATP>", status);
}

/*-----------------------------------------------------------------*/
/*! write the end OpenMath tag </OMATP>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeendOMATP(SCSCP_io* stream, SCSCP_status* status)
{
 return SCSCP_io_write(stream, "</OMATP>", status);
}

/*-----------------------------------------------------------------*/
/*! write the expression <OME><OMS cd="cdname" name="symbolname><OMSTR>buffer</OMSTR></OME>"
   directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
 @param cdname (in) cd name
 @param symbolname (in) symbol name
 @param buffer (in) message
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeOMEwithOMSOMSTR(SCSCP_io* stream, const char *cdname, const char*symbolname, 
                                  const char *buffer, SCSCP_status* status)
{
 int res = SCSCP_io_writebeginOME(stream, status);
 SCSCP_debugprint("SCSCP_io_writeOMEwithOMSOMSTR(%p, %s ,%s, %s,%p) - enter\n", stream, cdname, symbolname, buffer, status);
 
 res = res && SCSCP_io_writePairOMSOMSTR(stream, cdname, symbolname, buffer, status);
 res = res && SCSCP_io_writeendOME(stream, status);
 
 SCSCP_debugprint("SCSCP_io_writeOMEwithOMSOMSTR(%d) - returns %d\n", SCSCP_status_is(status), res);
 return res;
}

/*-----------------------------------------------------------------*/
/*! write the expression <OMS cd="cdname" name="symbolname><OMSTR>buffer</OMSTR>"
   directly to the SCSCP client or server 
 return 0 on error
 @param stream (inout) scscp stream
 @param status (inout) status error
 @param cdname (in) cd name
 @param symbolname (in) symbol name
 @param buffer (in) message
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writePairOMSOMSTR(SCSCP_io* stream, const char *cdname, const char*symbolname, 
                                  const char *buffer, SCSCP_status* status)
{
 int res;
 SCSCP_debugprint("SCSCP_io_writePairOMSOMSTR(%p, %s ,%s, %s,%p) - enter\n", stream, cdname, symbolname, buffer, status);
 
 res = SCSCP_io_writeOMS(stream, cdname, symbolname, status);
 res = res && SCSCP_io_writeOMSTR(stream, buffer, status);

 SCSCP_debugprint("SCSCP_io_writePairOMSOMSTR(%d) - returns %d\n", SCSCP_status_is(status), res);
 return res;
}

/*-----------------------------------------------------------------*/
/*! write the expression <OMS cd="cdname" name="symbolname><OMI>value</OMI>"
   directly to the SCSCP client or server 
 return 0 on error
 @param stream (inout) scscp stream
 @param status (inout) status error
 @param cdname (in) cd name
 @param symbolname (in) symbol name
 @param buffer (in) message
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writePairOMSOMIint(SCSCP_io* stream, const char *cdname, const char*symbolname, 
                               int value, SCSCP_status* status)
{
 int res;
 SCSCP_debugprint("SCSCP_io_writePairOMSOMIint(%p, %s ,%s, %d,%p) - enter\n", stream, cdname, symbolname, value, status);
 
 res = SCSCP_io_writeOMS(stream, cdname, symbolname, status);
 res = res && SCSCP_io_writeOMIint(stream, value, status);

 SCSCP_debugprint("SCSCP_io_writePairOMSOMIint(%d) - returns %d\n", SCSCP_status_is(status), res);
 return res;
}

/*-----------------------------------------------------------------*/
/*! write the expression <OMS cd="cdname" name="symbolname><OMI>value</OMI>"
   directly to the SCSCP client or server 
 return 0 on error
 @param stream (inout) scscp stream
 @param status (inout) status error
 @param cdname (in) cd name
 @param symbolname (in) symbol name
 @param buffer (in) message
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writePairOMSOMIsizet(SCSCP_io* stream, const char *cdname, const char*symbolname, 
                               size_t value, SCSCP_status* status)
{
 int res;
 SCSCP_debugprint("SCSCP_io_writePairOMSOMIsizet(%p, %s ,%s, %d,%p) - enter\n", stream, cdname, symbolname, value, status);
 
 res = SCSCP_io_writeOMS(stream, cdname, symbolname, status);
 res = res && SCSCP_io_writeOMIsizet(stream, value, status);

 SCSCP_debugprint("SCSCP_io_writePairOMSOMIsizet(%d) - returns %d\n", SCSCP_status_is(status), res);
 return res;
}


/*-----------------------------------------------------------------*/
/*! write the SCSCP tag <?scscp start ?>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeSCSCPSTART(SCSCP_io* stream, SCSCP_status* status)
{
 int res= SCSCP_io_write(stream, "<?scscp start ?>", status);
 return res && SCSCP_io_flush(stream,status);
}

/*-----------------------------------------------------------------*/
/*! write the SCSCP tag <?scscp end ?>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeSCSCPEND(SCSCP_io* stream, SCSCP_status* status)
{
 int res =  SCSCP_io_flush(stream,status);
 res &= SCSCP_io_write(stream, "<?scscp end ?>", status);
 res &= SCSCP_io_flush(stream,status);
 return res;
}

/*-----------------------------------------------------------------*/
/*! write the SCSCP tag <?scscp cancel ?>  directly to the SCSCP client or server 
 return 0 on error
 @param client (inout) scscp stream
 @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_writeSCSCPCANCEL(SCSCP_io* stream, SCSCP_status* status)
{
 int res =  SCSCP_io_flush(stream,status);
 res &= SCSCP_io_write(stream, "<?scscp cancel ?>", status);
 res &= SCSCP_io_flush(stream,status);
 return res;
}

/*-----------------------------------------------------------------*/
/*! checks if SCSCP stream is null and returns 0 on error
  @param client (in) object to be checked
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
static int SCSCP_io_checkNULL(SCSCP_io* stream, SCSCP_status* status)
{
   int res = 1;
   
   if (stream==NULL || *stream==NULL)
   {
     SCSCP_status_seterror(status,  SCSCP_STATUS_STREAMOBJECTNULL);
     res = 0;
   }
   return res;
}

/*-----------------------------------------------------------------*/
/*! \internal read one processing instruction (PI) of the stream.
  get the content of the PI from the stream
  Store the type of PI
  Store the content of PI
  @param client (in) object to be checked
  @param msgtype (out) type of returned message.
   could be SCSCP_msgtype_ProcedureCall, SCSCP_msgtype_Interrupt, in the current specification
  @param bufferPI (out) buffer will contain the PI. 
   It must contain at least SCSCP_PI_MAXLENBUFFER characters
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_ReadPI(SCSCP_io* stream, SCSCP_msgtype* msgtype, char *bufferPI, SCSCP_status* status)
{
 const char *PIinfo = "<?scscp info";
 const char *PIstart = "<?scscp start ?>";
 const char *PIquit = "<?scscp quit ";
 const char *PIterminate = "<?scscp terminate ";
 int res;
 int finish;
 
  SCSCP_debugprint("SCSCP_io_ReadPI() - enter\n");
  res = SCSCP_io_checkNULL(stream, status);
  if (res)
  {
    do
    {
      char *buffer = bufferPI;
      finish=1;
      /* read the type of PI */
      res = SCSCP_sc_readline(*stream, buffer, SCSCP_PI_MAXLENBUFFER, status);
      /* check the type of PI */
      if (res)
      {
       /* locate "<?" */
       buffer = strstr(buffer, "<?");
       if (buffer==NULL)
       {
         SCSCP_debugprint("unknown PI: '%s'\n", buffer);
         finish = 0;
       }
       else if (strncmp(buffer,PIinfo, strlen(PIinfo))==0)
       {
        SCSCP_debugprint("information PI : '%s'\n", buffer);
        finish = 0;
       }
       else if (strncmp(buffer,PIstart, strlen(PIstart))==0)
       {
         SCSCP_debugprint("start PI : '%s'\n", buffer);
         *msgtype = SCSCP_msgtype_ProcedureCall;
       }
       else if (strncmp(buffer,PIquit, strlen(PIquit))==0)
       {
         SCSCP_debugprint("quit PI : '%s'\n", buffer);
         res = 0;
         SCSCP_status_seterror(status, SCSCP_STATUS_RECVQUIT);
       }
       else if (strncmp(buffer,PIterminate, strlen(PIterminate))==0)
       {
         SCSCP_debugprint("interrupt PI : '%s'\n", buffer);
         *msgtype = SCSCP_msgtype_Interrupt;
       }
       else
       {
         SCSCP_debugprint("unknown PI: '%s'\n", buffer);
         finish = 0;
       }
      } 
     } while(res==1 && finish==0); 
   }
  SCSCP_debugprint("SCSCP_io_ReadPI(,%d,%d) - returns %d\n",(int)*msgtype, SCSCP_status_is(status), res);
  return res;
}


/*-----------------------------------------------------------------*/
/*! \internal check the xml document and return an iterator on the head of the document.
  @param stream (inout) stream
  @param iter (out) racine du document
  @param status (inout) status error
*/
/*-----------------------------------------------------------------*/
int SCSCP_io_readxmldoc(SCSCP_io *stream, SCSCP_xmlnodeptr* iter, SCSCP_status *status)
{
 int res =  SCSCP_sc_readxmldoc(*stream, status);
 if (res) *iter =  xmlDocGetRootElement((*stream)->m_doc);
 return res;
}

