/*-----------------------------------------------------------------*/
/*! 
  \file omlxmlnode.c
  \brief  xml node  implementation.
  \author  M. Gastineau 
           Astronomie et Systemes Dynamiques, IMCCE, CNRS, Observatoire de Paris. 

   Copyright, 2008, 2009, M. Gastineau, IMCCE-CNRS
   email of the author : gastineau@imcce.fr
*/
/*-----------------------------------------------------------------*/

/*-----------------------------------------------------------------*/
/* License  of this file :
This file is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute this file under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.

*/
/*-----------------------------------------------------------------*/


#if HAVE_CONFIG_H
#include "scscpconfig.h"
#endif /*HAVE_CONFIG_H*/
#include <stdio.h>
#if HAVE_STRING_H
#include <string.h>
#endif
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#include <errno.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "scscptypes.h"
#include "scscpinternal.h"
#include "scscpdebug.h"
#include "scscpfileclient.h"
#include "scscpfileserver.h"
#define __SCSCP_WITHIN_SCSCP
#include "scscp.h"

/***************************************************************************/
/* macros locales aux fichiers */
/***************************************************************************/
/*! conversion de xmlchar en char */
#define xmlChar2char(x) ((const char*)(x))

/*-----------------------------------------------------------------*/
/* variables locales */
/*-----------------------------------------------------------------*/
//! table of openmath object
#if 0
const struct 
{ 
	OML_inputstream::OMLtype type; 
	const char *name;
} 
s_tableindicename[]=
{
 { OML_inputstream::eOMLOMI, "OMI" },
 { OML_inputstream::eOMLOMS, "OMS" },
 { OML_inputstream::eOMLOMA, "OMA" },
 { OML_inputstream::eOMLOMV, "OMV" },
 { OML_inputstream::eOMLOMF, "OMF" },
 { OML_inputstream::eOMLOMSTR, "OMSTR" },
 { OML_inputstream::eOMLInvalid, "" }
};
#endif


/*----------------------------------------------------------------------------------------------*/
/*! return the next node after the current xml node.
  return NULL if curnode is the last node.
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlnodeptr SCSCP_xmlnode_getnext(SCSCP_xmlnodeptr curnode)
{
 return curnode->next;
}

/*----------------------------------------------------------------------------------------------*/
/*! return the name after the current xml node.
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
const char* SCSCP_xmlnode_getname(SCSCP_xmlnodeptr curnode)
{ 
 return xmlChar2char(curnode->name);
}


/*----------------------------------------------------------------------------------------------*/
/*! return the first child node the current xml node.
  return NULL if no child present
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlnodeptr SCSCP_xmlnode_getchild(SCSCP_xmlnodeptr curnode)
{ 
 return curnode->children;
}

/*----------------------------------------------------------------------------------------------*/
/*! read the content of the current node in the current xml node
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
const char * SCSCP_xmlnode_getcontent(SCSCP_xmlnodeptr curnode)
{ 
 return xmlChar2char(XML_GET_CONTENT(curnode));
}

/*----------------------------------------------------------------------------------------------*/
/*! read the first attribute of the current node
  return NULL if no attribute present
  @param curnode (in) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlattrptr SCSCP_xmlnode_getattr(SCSCP_xmlnodeptr curnode)
{ 
 return  curnode->properties;
}

/*-----------------------------------------------------------------*/
/*! return the associated value to the attributes specified by the name
  return NULL if not found
  @param iter (in) iterator on the xml doc
  @param p_pszName (in) name of the attribute
*/
/*-----------------------------------------------------------------*/
const char* SCSCP_xmlnode_findattr(SCSCP_xmlnodeptr iter, const char *p_pszName)
{
 const char* attr = NULL;
 const xmlChar* szType;
 struct _xmlAttr* pattAttribut=iter->properties;
 
 SCSCP_debugprint("OML_FindAttributes(%p,%s) -enter\n", iter, p_pszName);
 
  while (pattAttribut && attr==NULL)
  {
            szType=pattAttribut->name;
            if (strcmp(xmlChar2char(szType),p_pszName)==0)
 			{
 				attr = xmlChar2char(pattAttribut->children->content);
  			}
 			pattAttribut=pattAttribut->next;
  }

 SCSCP_debugprint("OML_FindAttributes() -returns %p\n", attr);
 return attr;
}

/*----------------------------------------------------------------------------------------------*/
/*! get the content of the attribute 
  return 0 on failure and non-zero on success
  @param attr (in) current attribute. mustn't be NULL 
  @param name (out) name of this attribute
  @param value (out) value of this attribute
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlattr_getvalue(SCSCP_xmlattrptr attr, const char **name, const char **value)
{ 
 *name = xmlChar2char(attr->name);
 *value = xmlChar2char(attr->children->content);
 return  1;
}

/*----------------------------------------------------------------------------------------------*/
/*!  return the next attribute after attr 
  return NULL if attr is the last attribute
  @param attr (in) current valid attribute. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
SCSCP_xmlattrptr SCSCP_xmlattr_getnext(SCSCP_xmlattrptr attr)
{ 
 return attr->next;
}



/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath Symbol and return its name and its cd.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm cdname (out) name of CD
  @param symbolname (out) name of the symbol
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMS(SCSCP_xmlnodeptr* curnode, const char **cdname, const char **symbolname)
{
  int res ;
  const char* szContent;
  const char* szType;
  
  SCSCP_xmlattrptr pattAttribut;

  SCSCP_debugprint("SCSCP_xmlnode_readOMS(%p,%p,%p) - enter\n", *curnode, cdname, symbolname);
 	
  *cdname = NULL;
  *symbolname = NULL;
  pattAttribut = SCSCP_xmlnode_getattr(*curnode);

  while (pattAttribut!=NULL)
  {
 			SCSCP_xmlattr_getvalue(pattAttribut, &szType, &szContent);
 			if (strcmp(szType,"cd")==0)
 			{
 				*cdname = szContent;
  			}
  			else if (strcmp(xmlChar2char(szType),"name")==0)
  			{
  				*symbolname = szContent;
 			}
 			pattAttribut=SCSCP_xmlattr_getnext(pattAttribut);
  }
  res = (*symbolname!=NULL)?1:0;
 
  *curnode = SCSCP_xmlnode_getnext(*curnode);
   
 SCSCP_debugprint("SCSCP_xmlnode_readOMS(%p,%s,%s) - returns %d\n", *curnode, *cdname, *symbolname,res);
 return res;
}

/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath integer and return its value.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMIstr(SCSCP_xmlnodeptr* curnode, const char **value)
{
  int res;
  *value = SCSCP_xmlnode_getcontent(SCSCP_xmlnode_getchild(*curnode));
  *curnode = SCSCP_xmlnode_getnext(*curnode);
  res =(*value!=NULL);
  SCSCP_debugprint("SCSCP_xmlnode_readOMIstr(%p,%s) - returns %d\n", *curnode, *value,res);
  return res;
}

/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath integer and return its value.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMIint(SCSCP_xmlnodeptr* curnode, int* value)
{
 const char *buffer;
 int res = SCSCP_xmlnode_readOMIstr(curnode, &buffer);
 if (res)
 {
  long lval;
  errno = 0;
  lval = strtol(buffer, (char **)NULL, 10);
  *value = (int)lval;
  res = (errno==0 && ((long)*value)==lval);
 }
 SCSCP_debugprint("SCSCP_xmlnode_readOMIint(%p,%d) - returns %d\n", *curnode, *value,res);
 return res;
}

/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath integer and return its value.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMIsizet(SCSCP_xmlnodeptr* curnode, size_t* value)
{
 const char *buffer;
 int res = SCSCP_xmlnode_readOMIstr(curnode, &buffer);
 if (res)
 {
  unsigned long long lval;
  errno = 0;
  lval = strtoull(buffer, (char **)NULL, 10);
  *value = (size_t)lval;
  res = (errno==0 && ((unsigned long long)*value)==lval);
 }
 SCSCP_debugprint("SCSCP_xmlnode_readOMIsizet(%p,%d) - returns %d\n", *curnode, *value,res);
 return res;
}


/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath error and return the optional arguments as a raw string.
  return 0  on error
  @param client (in) client session
  @param node (in) current valid node. mustn't be NULL
                   return the node just after the head symbol 
  @param cdname (out) name of the CD
  @param symbolname (out) name of the symbol
  @param value (out) openmath expression (after the 1st symbol). ust be freed by the user using free
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_sc_xmlnodereadOMEstr(SCSCP_socketclient* client, SCSCP_xmlnodeptr* curnode, const char **cdname, 
                             const char **symbolname, char **value, SCSCP_status *status)
{
 int res;
 SCSCP_xmlnodeptr child;
 SCSCP_debugprint("SCSCP_xmlnode_readOMEstr(%p, %p, %p,%p,%p) -enter\n", curnode, cdname, symbolname, value);
  
  child = SCSCP_xmlnode_getchild(*curnode);
  if (child==NULL)
  {
   res = 0;
   SCSCP_status_seterror(status, SCSCP_STATUS_OPENMATHNOTVALID);
  }
  else
  {
   res = SCSCP_xmlnode_readOMS(&child, cdname, symbolname);
   
   if (res)
   {
    *value = SCSCP_sc_getxmlnoderawstring(client, child, status);
   
    *curnode = SCSCP_xmlnode_getnext(*curnode);
    res = (*value!=NULL);
   }
  }
 
 SCSCP_debugprint("SCSCP_xmlnode_readOMEstr(,%s,%d) - returns %d\n", *value, SCSCP_status_is(status), res);
 return res; 
}


/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath floating-point  and return its value.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMFstr(SCSCP_xmlnodeptr* curnode, const char **value)
{
  int res;
  SCSCP_xmlnodeptr child = SCSCP_xmlnode_getchild(*curnode);
 
  *value = SCSCP_xmlnode_findattr(*curnode, "dec");
  if (*value==NULL)
  {
   *value = SCSCP_xmlnode_findattr(*curnode, "hex");
  }
  if (*value==NULL && child!=NULL)
  {
   *value = SCSCP_xmlnode_getcontent(child);
  }
  res = (*value!=NULL);
  if (res) *curnode = SCSCP_xmlnode_getnext(*curnode);
  
 SCSCP_debugprint("SCSCP_xmlnode_readOMFstr(%p, %s) - returns %d\n", *curnode, *value, res);
 return res;
}

/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath string  and return its value.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMSTR(SCSCP_xmlnodeptr* curnode, const char **value)
{
 int res;
 SCSCP_xmlnodeptr child = SCSCP_xmlnode_getchild(*curnode);
 
  *value = (child==NULL)? NULL: SCSCP_xmlnode_getcontent(child);
  if (*value==NULL || strlen(*value)==0) 
     res = 0;
  else 
  {
   *curnode = SCSCP_xmlnode_getnext(*curnode);
   res=1;
  }
  
 SCSCP_debugprint("SCSCP_xmlnode_readOMSTR(%p, %s) - returns %d\n", *curnode, *value, res);
 return res;
}


/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath reference and return its value.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) reference name
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMR(SCSCP_xmlnodeptr* curnode, const char **value)
{
 int res;
 SCSCP_debugprint("SCSCP_xmlnode_readOMR(%p, %p) -enter\n", *curnode, value);

  *value = SCSCP_xmlnode_findattr(*curnode, "xref");
  if (*value==NULL) 
  {
   res=0;
  }
  else
  {
   *curnode = SCSCP_xmlnode_getnext(*curnode);
   res = 1;
  }
 
 SCSCP_debugprint("SCSCP_xmlnode_readOMR(,%s) - returns %d\n", *value, res);
 return res;
}


/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath variable  and return its value.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readOMV(SCSCP_xmlnodeptr* curnode, const char **value)
{
 int res;
 SCSCP_xmlnodeptr child = SCSCP_xmlnode_getchild(*curnode);
 
  *value = SCSCP_xmlnode_findattr(*curnode, "name");
  if (*value==NULL && child!=NULL)
  {
   *value = SCSCP_xmlnode_getcontent(child);
  }
  res = (*value!=NULL);
  
  *curnode = SCSCP_xmlnode_getnext(*curnode);
  
 SCSCP_debugprint("SCSCP_xmlnode_readOMV(%p, %s) - returns %d\n", *curnode, *value, res);
 return res;
}


/*----------------------------------------------------------------------------------------------*/
/*! find the first <OMOBJ>, look in child node
  return 0  on error
  @param node (inout) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readbeginOMOBJ(SCSCP_xmlnodeptr* curnode)
{
 return SCSCP_xmlnode_movetotag(curnode, "OMOBJ", 1,1);
}

/*----------------------------------------------------------------------------------------------*/
/*! find the first <OMA>, look in child node
  return 0  on error
  @param node (inout) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readbeginOMA(SCSCP_xmlnodeptr* curnode)
{
 return SCSCP_xmlnode_movetotag(curnode, "OMA", 1,1);
}

/*----------------------------------------------------------------------------------------------*/
/*! find the first <OMATTR>, look in child node
  return 0  on error
  @param node (inout) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readbeginOMATTR(SCSCP_xmlnodeptr* curnode)
{
 return SCSCP_xmlnode_movetotag(curnode, "OMATTR", 1,1);
}

/*----------------------------------------------------------------------------------------------*/
/*! find the first <OMATP>, look in child node
  return 0  on error
  @param node (inout) current valid node. mustn't be NULL 
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readbeginOMATP(SCSCP_xmlnodeptr* curnode)
{
 return SCSCP_xmlnode_movetotag(curnode, "OMATP", 1,1);
}

/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath Symbol, check the symbol name with symbolname and the CD anme 
     and return the next OpenMath string.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readpairOMSOMSTR(SCSCP_xmlnodeptr* curnode, const char *cdname, const char *symbolname, const char **buffer)
{
 int res = SCSCP_xmlnode_readcheckOMS(curnode, cdname, symbolname);
 if (res) res = SCSCP_xmlnode_readOMSTR(curnode, buffer);
 return res;
}

/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath Symbol, check the symbol name with symbolname and the CD anme 
     and return the next OpenMath integer.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readpairOMSOMIsizet(SCSCP_xmlnodeptr* curnode, const char *cdname, const char *symbolname, size_t *value)
{
 int res = SCSCP_xmlnode_readcheckOMS(curnode, cdname, symbolname);
 if (res) res = SCSCP_xmlnode_readOMIsizet(curnode, value);
 return res;
}

/*----------------------------------------------------------------------------------------------*/
/*!  read an OpenMath Symbol, check the symbol name with symbolname and the CD anme 
     and return the next OpenMath integer.
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readpairOMSOMIint(SCSCP_xmlnodeptr* curnode, const char *cdname, const char *symbolname, int *value)
{
 int res = SCSCP_xmlnode_readcheckOMS(curnode, cdname, symbolname);
 if (res) res = SCSCP_xmlnode_readOMIint(curnode, value);
 return res;
}

/*----------------------------------------------------------------------------------------------*/
/*!   move the current node pointer to the specified symbol
     read an OpenMath Symbol, check the symbol name with symbolname and the CD anme 
  return 0  on error
  @param node (in) current valid node. mustn't be NULL 
  @parm value (out) value of the integer
*/
/*----------------------------------------------------------------------------------------------*/
int SCSCP_xmlnode_readcheckOMS(SCSCP_xmlnodeptr* curnode, const char *cdname, const char *symbolname)
{
 int res = 1;
 int bfound = 0;
 SCSCP_debugprint("SCSCP_xmlnode_readcheckOMS(%p,%s, %s ) - enter\n", curnode, cdname, symbolname);

 while (!bfound && res)
 {
  if (SCSCP_xmlnode_movetotag(curnode, "OMS", 0, 0))
  {
    /* check cdname and name */
    const char *cdread, *symbolread;
    if ((res=SCSCP_xmlnode_readOMS(curnode, &cdread, &symbolread))==1)
    {
     bfound = (strcmp(cdname,cdread)==0 && strcmp(symbolname, symbolread)==0);
    }
  } 
  else { break; /* exit from loop */ }
 }
 
 res = res & bfound;
 
 SCSCP_debugprint("SCSCP_xmlnode_readcheckOMS(%p) - returns %d\n", curnode, res);
 return res;
}

 /*-----------------------------------------------------------------*/
/*!move the current node pointer to the first specified item
 return true on success
  @param curnode (inout) iterator on the xml doc
  @param string (in) string to find
  @param children (in) = true => look only in the children
  @param gotochildnode (in) = 0 => return the looking node
                           = 1 =>  return the first child node
*/
/*-----------------------------------------------------------------*/
int SCSCP_xmlnode_movetotag(SCSCP_xmlnodeptr* curnode, const char *string, int children, int gotochildnode)
{
  int res = 0;
  SCSCP_xmlnodeptr itertmp = *curnode;
  SCSCP_xmlnodeptr iter = *curnode;
  
  SCSCP_debugprint("SCSCP_xmlnode_movetotag(%p,%s,%d) - enter\n", *curnode, string, children);

  if (children)
  {
      if (iter)
      {
       if (iter->type==XML_TEXT_NODE)
       { // cas de texte entre les noeuds
        SCSCP_xmlnodeptr next = SCSCP_xmlnode_getnext(iter);
        res = SCSCP_xmlnode_movetotag(&next, string, children, gotochildnode);
       }
       else if (iter->type == XML_ELEMENT_NODE) 
       {
        SCSCP_debugprint("SCSCP_xmlnode_movetotag() - node %s type XML_ELEMENT_NODE \n", iter->name);
        res = (xmlStrcmp((const xmlChar *)string, iter->name)==0);
       }
       if (!res) 
       {
        itertmp = SCSCP_xmlnode_getchild(iter);
        res = SCSCP_xmlnode_movetotag(&itertmp, string, children, gotochildnode);
        if (res) *curnode = itertmp;
       }
      }
  }
  else
  {
      while (iter!=NULL && !res)
      {
       if (iter->type == XML_ELEMENT_NODE) 
       {
        SCSCP_debugprint("SCSCP_xmlnode_movetotag() - node %s type XML_ELEMENT_NODE \n", iter->name);
        res = (xmlStrcmp((const xmlChar *)string, iter->name)==0);
       }
       if (!res) iter = SCSCP_xmlnode_getnext(iter);
      }
      *curnode = iter;
  }      
 
 if (res && gotochildnode) *curnode = SCSCP_xmlnode_getchild(*curnode);
 
 SCSCP_debugprint("SCSCP_xmlnode_movetotag(%p,) - returns %d\n", *curnode, res);
 return res;
}


#if 0
/********************************************************/
/*!
 return the indice from the name of the planet
 @param name (in) name of the planet
*/
/********************************************************/
OML_inputstream::OMLtype OML_inputstream::GetTypeFromName(const_OMLchar name) const
{
 OMLtype indice = eOMLInvalid;
 size_t cpt;
 
 for(cpt=0; cpt<sizeof(s_tableindicename)/sizeof(s_tableindicename[0]); cpt++)
 {
  if (strcmp(s_tableindicename[cpt].name, name)==0) 
  {
   indice=s_tableindicename[cpt].type;
   break; /* exit the loop */
  }
 }
 
 /* not found => generate fatal error */
 return indice;
}

/*-----------------------------------------------------------------*/
/*! return the type of the object pointed by iter
*/
/*-----------------------------------------------------------------*/
OML_inputstream::OMLtype OML_inputstream::GetOMtype(const iterator& iter) const
{
  OMLtype etype = eOMLInvalid;
  if (iter->type == XML_ELEMENT_NODE) 
  {
   etype = GetTypeFromName(xmlChar2char(iter->name));
  }
  return etype;
}

#endif

